"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var TokenKind;
(function (TokenKind) {
    TokenKind["Unknown"] = "Unknown";
    TokenKind["EndOfFileToken"] = "EndOfFileToken";
    TokenKind["NameToken"] = "NameToken";
    TokenKind["ColonToken"] = "ColonToken";
    TokenKind["MinusToken"] = "MinusToken";
    TokenKind["DoubleQuoteToken"] = "DoubleQuoteToken";
    TokenKind["SingleQuoteToken"] = "SingleQuoteToken";
    TokenKind["OtherToken"] = "OtherToken";
    TokenKind["MissingToken"] = "MissingToken";
    TokenKind["SkippedToken"] = "SkippedToken";
    // Used for "MissingToken" values
    TokenKind["Expression"] = "Expression";
    TokenKind["Value"] = "Value";
})(TokenKind = exports.TokenKind || (exports.TokenKind = {}));
// Based on https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Token.php
class Token {
    constructor(kind, fullStart, start = fullStart, length = 0) {
        this.kind = kind;
        this.fullStart = fullStart;
        this.start = start;
        this.length = length;
    }
    text(text) {
        return text.substring(this.start, this.fullStart + this.length);
    }
    fullText(text) {
        return text.substring(this.fullStart, this.fullStart + this.length);
    }
}
exports.Token = Token;
class MissingToken extends Token {
}
exports.MissingToken = MissingToken;
const textToToken = {
    ":": TokenKind.ColonToken,
    "'": TokenKind.SingleQuoteToken,
    '"': TokenKind.DoubleQuoteToken,
    "-": TokenKind.MinusToken
};
class QueryNode {
    constructor(children = []) {
        this.children = children;
    }
    getChildren() {
        return this.children;
    }
}
exports.QueryNode = QueryNode;
class NegateNode {
    constructor(minusToken, parent) {
        this.minusToken = minusToken;
        this.parent = parent;
    }
    getChildren() {
        return [this.minusToken, this.expression];
    }
}
exports.NegateNode = NegateNode;
class NameNode {
    constructor(name, parent) {
        this.name = name;
        this.parent = parent;
    }
    getValue(text) {
        return this.name.text(text);
    }
    getChildren() {
        return [this.name];
    }
}
exports.NameNode = NameNode;
class StringLiteralNode {
    constructor(parent) {
        this.parent = parent;
    }
    getValue(text) {
        return this.text.map(token => token.fullText(text)).join("");
    }
    getChildren() {
        return [this.openQuote, ...this.text, this.closeQuote];
    }
}
exports.StringLiteralNode = StringLiteralNode;
class FieldNode {
    constructor(parent) {
        this.parent = parent;
    }
    getChildren() {
        return [this.name, this.colon, this.expression];
    }
}
exports.FieldNode = FieldNode;
class InsertedToken {
    constructor(text) {
        this.text = text;
    }
}
exports.InsertedToken = InsertedToken;
class SkippedNode {
    constructor(token) {
        this.token = token;
    }
    getChildren() {
        return [this.token];
    }
}
exports.SkippedNode = SkippedNode;
class Scanner {
    constructor(text) {
        this.text = text;
        this.pos = 0;
        this.endOfFilePos = this.text.length;
    }
    scanNextToken() {
        const fullStart = this.pos;
        while (true) {
            const start = this.pos;
            if (this.pos >= this.endOfFilePos) {
                return new Token(TokenKind.EndOfFileToken, fullStart, start, this.pos - fullStart);
            }
            const char = this.text[this.pos];
            // Skip over whitespace, it contributes to "before trivia"
            if (char.match(/\s/) !== null) {
                this.pos++;
                continue;
            }
            // Check single-char to token mapping
            const kind = textToToken[char];
            if (kind !== undefined) {
                this.pos++;
                return new Token(kind, fullStart, start, this.pos - fullStart);
            }
            let token;
            // Check for a name
            if ((token = this.tryScanName(fullStart, start)) !== null) {
                return token;
            }
            // Capture anything else as an OtherToken
            if ((token = this.tryScanOther(fullStart, start)) !== null) {
                return token;
            }
            throw new Error(`Unexpected character: ${char} charCodeAt(0)=${char.charCodeAt(0)}`);
        }
    }
    tryScanName(fullStart, start) {
        while (this.pos < this.endOfFilePos) {
            const char = this.text[this.pos];
            if (char.match(/\w/u) !== null) {
                this.pos++;
            }
            else {
                break;
            }
        }
        if (this.pos > start) {
            return new Token(TokenKind.NameToken, fullStart, start, this.pos - fullStart);
        }
        else {
            return null;
        }
    }
    tryScanOther(fullStart, start) {
        if (this.pos < this.endOfFilePos) {
            this.pos++;
            return new Token(TokenKind.OtherToken, fullStart, start, this.pos - fullStart);
        }
        return null;
    }
    lookahead(tokenKind) {
        const originalPos = this.pos;
        const result = this.scanNextToken().kind === tokenKind;
        this.pos = originalPos;
        return result;
    }
}
function parse(text) {
    const query = new QueryNode();
    const lexer = new Scanner(text);
    let token = lexer.scanNextToken();
    function advanceToken() {
        token = lexer.scanNextToken();
    }
    function eat(tokenKind) {
        if (token.kind === tokenKind) {
            const eaten = token;
            advanceToken();
            return eaten;
        }
        return new MissingToken(tokenKind, token.fullStart);
    }
    function parseExpression(parentNode) {
        switch (token.kind) {
            case TokenKind.MinusToken: {
                const negateNode = new NegateNode(token, parentNode);
                advanceToken();
                negateNode.expression = parsePositiveExpression(negateNode);
                return negateNode;
            }
            default: {
                return parsePositiveExpression(parentNode);
            }
        }
    }
    function parsePositiveExpression(parentNode) {
        if (token.kind === TokenKind.NameToken) {
            if (lexer.lookahead(TokenKind.ColonToken)) {
                return parseField(parentNode);
            }
        }
        return parseValue(parentNode);
    }
    function parseValue(parentNode) {
        switch (token.kind) {
            case TokenKind.NameToken: {
                const node = new NameNode(token, parentNode);
                advanceToken();
                return node;
            }
            case TokenKind.SingleQuoteToken:
            case TokenKind.DoubleQuoteToken: {
                const quoteTokenKind = token.kind;
                const stringLiteral = new StringLiteralNode(parentNode);
                stringLiteral.openQuote = eat(quoteTokenKind);
                stringLiteral.text = [];
                while (token.kind !== quoteTokenKind && token.kind !== TokenKind.EndOfFileToken) {
                    stringLiteral.text.push(token);
                    advanceToken();
                }
                stringLiteral.closeQuote = eat(quoteTokenKind);
                return stringLiteral;
            }
            default: {
                return new MissingToken(TokenKind.Value, token.fullStart);
            }
        }
    }
    function parseField(parentNode) {
        const fieldNode = new FieldNode(parentNode);
        fieldNode.name = token;
        advanceToken();
        fieldNode.colon = eat(TokenKind.ColonToken);
        fieldNode.expression = parseValue(fieldNode);
        return fieldNode;
    }
    // Step through each token and attempt to parse.
    while (token.kind !== TokenKind.EndOfFileToken) {
        const startToken = token;
        switch (token.kind) {
            case TokenKind.ColonToken:
            case TokenKind.OtherToken: {
                query.children.push(new SkippedNode(token));
                advanceToken();
                break;
            }
            default: {
                query.children.push(parseExpression(query));
            }
        }
        if (token === startToken) {
            throw new Error(`Failed to parse token ${token.kind}`);
        }
    }
    // Capture the EndOfFileToken.
    query.children.push(token);
    return query;
}
exports.parse = parse;
function getFullText(node, text) {
    if (node instanceof InsertedToken) {
        return node.text;
    }
    const children = node instanceof Token ? [node] : node.getChildren();
    return children
        .map(nodeOrToken => nodeOrToken instanceof InsertedToken
        ? nodeOrToken.text
        : nodeOrToken instanceof Token
            ? nodeOrToken.fullText(text)
            : getFullText(nodeOrToken, text))
        .join("");
}
exports.getFullText = getFullText;
function toTsQuery(node, text) {
    if (node instanceof Token) {
        return node.text(text);
    }
    else if (node instanceof QueryNode) {
        const terms = [];
        for (const child of node.children) {
            const term = toTsQuery(child, text);
            if (term.length > 0) {
                terms.push(term);
            }
        }
        return terms.join(" & ");
    }
    else if (node instanceof NameNode) {
        return node.name.text(text);
    }
    else if (node instanceof NegateNode) {
        const exprText = toTsQuery(node.expression, text);
        return exprText.length > 0 ? `! ${exprText}` : "";
    }
    return "";
}
exports.toTsQuery = toTsQuery;
function isUuid(text) {
    return text.match(/^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$/) !== null;
}
function toPredicate(node, text) {
    const predicates = [];
    function parseTagField(node) {
        const { name, expression } = node;
        if (name.text(text) === "tag") {
            if (expression instanceof NameNode) {
                const value = expression.name.text(text);
                return isUuid(value) ? { type: PredicateKind.TagId, id: value } : { type: PredicateKind.TagTitleFuzzy, title: value };
            }
            else if (expression instanceof StringLiteralNode) {
                return { type: PredicateKind.TagTitleExact, title: expression.text.map(token => token.fullText(text)).join("") };
            }
        }
        return;
    }
    function parseTypeField(node) {
        const { name, expression } = node;
        if (name.text(text) === "type") {
            const value = expression instanceof NameNode || expression instanceof StringLiteralNode ? expression.getValue(text) : "";
            if (value === "tag" || value === "insight" || value === "highlight" || value === "note") {
                return { type: PredicateKind.ResultType, resultType: value };
            }
        }
        return;
    }
    function parseProjectField(node) {
        const { name, expression } = node;
        if (name.text(text) === "project") {
            if (expression instanceof NameNode) {
                const value = expression.name.text(text);
                return isUuid(value)
                    ? { type: PredicateKind.ProjectId, id: value }
                    : { type: PredicateKind.ProjectTitleFuzzy, title: value };
            }
            else if (expression instanceof StringLiteralNode) {
                return { type: PredicateKind.ProjectTitleExact, title: expression.text.map(token => token.text(text)).join("") };
            }
        }
        return;
    }
    function compile(child) {
        let value;
        if (child instanceof FieldNode) {
            if ((value = parseTagField(child)) !== undefined ||
                (value = parseProjectField(child)) !== undefined ||
                (value = parseTypeField(child)) !== undefined) {
                return value;
            }
        }
        else if (child instanceof NegateNode) {
            const predicate = compile(child.expression);
            if (predicate !== undefined) {
                return { type: PredicateKind.Not, predicate: predicate };
            }
        }
        else {
            const tsquery = toTsQuery(child, text);
            if (tsquery.length > 0) {
                return {
                    type: PredicateKind.TsQuery,
                    tsQuery: tsquery
                };
            }
        }
        return;
    }
    for (const child of node.children) {
        const predicate = compile(child);
        if (predicate !== undefined) {
            predicates.push(predicate);
        }
    }
    return predicates.length > 0
        ? predicates.slice(1).reduce((accum, curr) => ({ type: PredicateKind.And, left: accum, right: curr }), predicates[0])
        : { type: PredicateKind.NoOp };
}
exports.toPredicate = toPredicate;
var PredicateKind;
(function (PredicateKind) {
    PredicateKind["And"] = "And";
    PredicateKind["NoOp"] = "NoOp";
    PredicateKind["Or"] = "Or";
    PredicateKind["ProjectId"] = "ProjectId";
    PredicateKind["ProjectTitleExact"] = "ProjectTitleExact";
    PredicateKind["ProjectTitleFuzzy"] = "ProjectTitleFuzzy";
    PredicateKind["TagId"] = "TagId";
    PredicateKind["TagTitleExact"] = "TagTitleExact";
    PredicateKind["TagTitleFuzzy"] = "TagTitleFuzzy";
    PredicateKind["TsQuery"] = "TsQuery";
    PredicateKind["Not"] = "Not";
    PredicateKind["ResultType"] = "ResultType";
})(PredicateKind = exports.PredicateKind || (exports.PredicateKind = {}));
function refactorTypeTo(node, text, type) {
    const children = [];
    for (const child of node.children) {
        let fieldCandidate = child;
        if (child instanceof NegateNode) {
            fieldCandidate = child.expression;
        }
        if (fieldCandidate instanceof FieldNode) {
            if (fieldCandidate.name.text(text) === "type") {
                continue;
            }
        }
        children.push(child);
    }
    if (type !== null) {
        // If the node that will follow the type:___ field starts with
        // non-whitespace, insert some whitespace to separate it from the `type:___`
        // field.
        if (getFullText(children[0], text).match(/^[^\s]/) !== null) {
            children.splice(0, 0, new InsertedToken(" "));
        }
        children.splice(0, 0, new InsertedToken(`type:${type}`));
    }
    return new QueryNode(children);
}
exports.refactorTypeTo = refactorTypeTo;
//# sourceMappingURL=index.js.map