export enum TokenKind {
  Unknown = "Unknown",
  EndOfFileToken = "EndOfFileToken",

  NameToken = "NameToken",
  ColonToken = "ColonToken",
  MinusToken = "MinusToken",
  DoubleQuoteToken = "DoubleQuoteToken",
  SingleQuoteToken = "SingleQuoteToken",
  OtherToken = "OtherToken",
  MissingToken = "MissingToken",
  SkippedToken = "SkippedToken",

  // Used for "MissingToken" values
  Expression = "Expression",
  Value = "Value"
}

// Based on https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Token.php
export class Token {
  constructor(
    public readonly kind: TokenKind,
    public readonly fullStart: number,
    public readonly start: number = fullStart,
    public readonly length: number = 0
  ) {}

  public text(text: string) {
    return text.substring(this.start, this.fullStart + this.length);
  }

  public fullText(text: string) {
    return text.substring(this.fullStart, this.fullStart + this.length);
  }
}

export class MissingToken extends Token {}

export type TreeNode = Token | Node | InsertedToken;

// Based on https://github.com/Microsoft/tolerant-php-parser/blob/master/src/Node.php
interface IGetChildren {
  getChildren(): ReadonlyArray<TreeNode>;
}

const textToToken: { [char: string]: TokenKind | undefined } = {
  ":": TokenKind.ColonToken,
  "'": TokenKind.SingleQuoteToken,
  '"': TokenKind.DoubleQuoteToken,
  "-": TokenKind.MinusToken
};

export class QueryNode implements IGetChildren {
  constructor(public readonly children: Array<TreeNode> = []) {}

  getChildren() {
    return this.children;
  }
}

export class NegateNode implements IGetChildren {
  public expression!: Node | MissingToken;

  constructor(private readonly minusToken: Token, public readonly parent: Node) {}

  getChildren() {
    return [this.minusToken, this.expression];
  }
}

export class NameNode implements IGetChildren {
  constructor(public readonly name: Token, public readonly parent: Node) {}

  getValue(text: string) {
    return this.name.text(text);
  }

  getChildren() {
    return [this.name];
  }
}

export class StringLiteralNode implements IGetChildren {
  public openQuote!: Token;
  public text!: Token[];
  public closeQuote!: Token;

  constructor(public readonly parent: IGetChildren) {}

  getValue(text: string) {
    return this.text.map(token => token.fullText(text)).join("");
  }

  getChildren() {
    return [this.openQuote, ...this.text, this.closeQuote];
  }
}

export class FieldNode implements IGetChildren {
  public name!: Token;
  public colon!: Token;
  public expression!: Node | MissingToken;

  constructor(public readonly parent: IGetChildren) {}

  getChildren() {
    return [this.name, this.colon, this.expression];
  }
}

export class InsertedToken {
  constructor(public readonly text: string) {}
}

export class SkippedNode implements IGetChildren {
  constructor(public readonly token: Token) {}

  getChildren() {
    return [this.token];
  }
}

export type Node = FieldNode | NameNode | NegateNode | QueryNode | SkippedNode | StringLiteralNode;

class Scanner {
  private pos = 0;
  private readonly endOfFilePos = this.text.length;

  constructor(private readonly text: string) {}

  public scanNextToken(): Token {
    const fullStart = this.pos;

    while (true) {
      const start = this.pos;

      if (this.pos >= this.endOfFilePos) {
        return new Token(TokenKind.EndOfFileToken, fullStart, start, this.pos - fullStart);
      }

      const char = this.text[this.pos];

      // Skip over whitespace, it contributes to "before trivia"
      if (char.match(/\s/) !== null) {
        this.pos++;
        continue;
      }

      // Check single-char to token mapping
      const kind = textToToken[char];
      if (kind !== undefined) {
        this.pos++;
        return new Token(kind, fullStart, start, this.pos - fullStart);
      }

      let token;

      // Check for a name
      if ((token = this.tryScanName(fullStart, start)) !== null) {
        return token;
      }

      // Capture anything else as an OtherToken
      if ((token = this.tryScanOther(fullStart, start)) !== null) {
        return token;
      }

      throw new Error(`Unexpected character: ${char} charCodeAt(0)=${char.charCodeAt(0)}`);
    }
  }

  private tryScanName(fullStart: number, start: number) {
    while (this.pos < this.endOfFilePos) {
      const char = this.text[this.pos];
      if (char.match(/\w/u) !== null) {
        this.pos++;
      } else {
        break;
      }
    }

    if (this.pos > start) {
      return new Token(TokenKind.NameToken, fullStart, start, this.pos - fullStart);
    } else {
      return null;
    }
  }

  private tryScanOther(fullStart: number, start: number) {
    if (this.pos < this.endOfFilePos) {
      this.pos++;
      return new Token(TokenKind.OtherToken, fullStart, start, this.pos - fullStart);
    }
    return null;
  }

  public lookahead(tokenKind: TokenKind): boolean {
    const originalPos = this.pos;
    const result = this.scanNextToken().kind === tokenKind;
    this.pos = originalPos;
    return result;
  }
}

export function parse(text: string): QueryNode {
  const query = new QueryNode();

  const lexer = new Scanner(text);
  let token = lexer.scanNextToken();

  function advanceToken() {
    token = lexer.scanNextToken();
  }

  function eat(tokenKind: TokenKind): MissingToken | Token {
    if (token.kind === tokenKind) {
      const eaten = token;
      advanceToken();
      return eaten;
    }

    return new MissingToken(tokenKind, token.fullStart);
  }

  function parseExpression(parentNode: Node) {
    switch (token.kind) {
      case TokenKind.MinusToken: {
        const negateNode = new NegateNode(token, parentNode);
        advanceToken();
        negateNode.expression = parsePositiveExpression(negateNode);
        return negateNode;
      }

      default: {
        return parsePositiveExpression(parentNode);
      }
    }
  }

  function parsePositiveExpression(parentNode: Node): Node | MissingToken {
    if (token.kind === TokenKind.NameToken) {
      if (lexer.lookahead(TokenKind.ColonToken)) {
        return parseField(parentNode);
      }
    }

    return parseValue(parentNode);
  }

  function parseValue(parentNode: Node): Node | MissingToken {
    switch (token.kind) {
      case TokenKind.NameToken: {
        const node = new NameNode(token, parentNode);
        advanceToken();
        return node;
      }

      case TokenKind.SingleQuoteToken:
      case TokenKind.DoubleQuoteToken: {
        const quoteTokenKind = token.kind;
        const stringLiteral = new StringLiteralNode(parentNode);
        stringLiteral.openQuote = eat(quoteTokenKind);
        stringLiteral.text = [];
        while (token.kind !== quoteTokenKind && (token as Token).kind !== TokenKind.EndOfFileToken) {
          stringLiteral.text.push(token);
          advanceToken();
        }
        stringLiteral.closeQuote = eat(quoteTokenKind);
        return stringLiteral;
      }

      default: {
        return new MissingToken(TokenKind.Value, token.fullStart);
      }
    }
  }

  function parseField(parentNode: Node): Node | MissingToken {
    const fieldNode = new FieldNode(parentNode);
    fieldNode.name = token;
    advanceToken();
    fieldNode.colon = eat(TokenKind.ColonToken);
    fieldNode.expression = parseValue(fieldNode);
    return fieldNode;
  }

  // Step through each token and attempt to parse.
  while (token.kind !== TokenKind.EndOfFileToken) {
    const startToken = token;
    switch (token.kind) {
      case TokenKind.ColonToken:
      case TokenKind.OtherToken: {
        query.children.push(new SkippedNode(token));
        advanceToken();
        break;
      }

      default: {
        query.children.push(parseExpression(query));
      }
    }

    if (token === startToken) {
      throw new Error(`Failed to parse token ${token.kind}`);
    }
  }

  // Capture the EndOfFileToken.
  query.children.push(token);

  return query;
}

export function getFullText(node: TreeNode, text: string): string {
  if (node instanceof InsertedToken) {
    return node.text;
  }

  const children = node instanceof Token ? [node] : node.getChildren();
  return children
    .map(
      nodeOrToken =>
        nodeOrToken instanceof InsertedToken
          ? nodeOrToken.text
          : nodeOrToken instanceof Token
            ? nodeOrToken.fullText(text)
            : getFullText(nodeOrToken, text)
    )
    .join("");
}

export function toTsQuery(node: TreeNode, text: string): string {
  if (node instanceof Token) {
    return node.text(text);
  } else if (node instanceof QueryNode) {
    const terms = [];
    for (const child of node.children) {
      const term = toTsQuery(child, text);
      if (term.length > 0) {
        terms.push(term);
      }
    }
    return terms.join(" & ");
  } else if (node instanceof NameNode) {
    return node.name.text(text);
  } else if (node instanceof NegateNode) {
    const exprText = toTsQuery(node.expression, text);
    return exprText.length > 0 ? `! ${exprText}` : "";
  }
  return "";
}

function isUuid(text: string) {
  return text.match(/^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$/) !== null;
}

export function toPredicate(node: QueryNode, text: string): Predicate {
  const predicates: Predicate[] = [];

  function parseTagField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (name.text(text) === "tag") {
      if (expression instanceof NameNode) {
        const value = expression.name.text(text);
        return isUuid(value) ? { type: PredicateKind.TagId, id: value } : { type: PredicateKind.TagTitleFuzzy, title: value };
      } else if (expression instanceof StringLiteralNode) {
        return { type: PredicateKind.TagTitleExact, title: expression.text.map(token => token.fullText(text)).join("") };
      }
    }
    return;
  }

  function parseTypeField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (name.text(text) === "type") {
      const value = expression instanceof NameNode || expression instanceof StringLiteralNode ? expression.getValue(text) : "";

      if (value === "tag" || value === "insight" || value === "highlight" || value === "note") {
        return { type: PredicateKind.ResultType, resultType: value };
      }
    }
    return;
  }

  function parseProjectField(node: FieldNode): Predicate | undefined {
    const { name, expression } = node;
    if (name.text(text) === "project") {
      if (expression instanceof NameNode) {
        const value = expression.name.text(text);
        return isUuid(value)
          ? { type: PredicateKind.ProjectId, id: value }
          : { type: PredicateKind.ProjectTitleFuzzy, title: value };
      } else if (expression instanceof StringLiteralNode) {
        return { type: PredicateKind.ProjectTitleExact, title: expression.text.map(token => token.text(text)).join("") };
      }
    }
    return;
  }

  function compile(child: TreeNode): Predicate | undefined {
    let value;
    if (child instanceof FieldNode) {
      if (
        (value = parseTagField(child)) !== undefined ||
        (value = parseProjectField(child)) !== undefined ||
        (value = parseTypeField(child)) !== undefined
      ) {
        return value;
      }
    } else if (child instanceof NegateNode) {
      const predicate = compile(child.expression);
      if (predicate !== undefined) {
        return { type: PredicateKind.Not, predicate: predicate };
      }
    } else {
      const tsquery = toTsQuery(child, text);
      if (tsquery.length > 0) {
        return {
          type: PredicateKind.TsQuery,
          tsQuery: tsquery
        };
      }
    }
    return;
  }

  for (const child of node.children) {
    const predicate = compile(child);
    if (predicate !== undefined) {
      predicates.push(predicate);
    }
  }

  return predicates.length > 0
    ? predicates.slice(1).reduce((accum, curr) => ({ type: PredicateKind.And, left: accum, right: curr }), predicates[0])
    : { type: PredicateKind.NoOp };
}

export enum PredicateKind {
  And = "And",
  NoOp = "NoOp",
  Or = "Or",
  ProjectId = "ProjectId",
  ProjectTitleExact = "ProjectTitleExact",
  ProjectTitleFuzzy = "ProjectTitleFuzzy",
  TagId = "TagId",
  TagTitleExact = "TagTitleExact",
  TagTitleFuzzy = "TagTitleFuzzy",
  TsQuery = "TsQuery",
  Not = "Not",
  ResultType = "ResultType"
}

export interface AndPredicate {
  readonly type: PredicateKind.And;
  readonly left: Predicate;
  readonly right: Predicate;
}

export interface OrPredicate {
  readonly type: PredicateKind.Or;
  readonly left: Predicate;
  readonly right: Predicate;
}

export interface ProjectTitleExactPredicate {
  readonly type: PredicateKind.ProjectTitleExact;
  readonly title: string;
}

export interface ProjectTitleFuzzyPredicate {
  readonly type: PredicateKind.ProjectTitleFuzzy;
  readonly title: string;
}

export interface ProjectIdPredicate {
  readonly type: PredicateKind.ProjectId;
  readonly id: string;
}

export interface TagIdPredicate {
  readonly type: PredicateKind.TagId;
  readonly id: string;
}

export interface TagTitleFuzzyPredicate {
  readonly type: PredicateKind.TagTitleFuzzy;
  readonly title: string;
}

export interface TagTitleExactPredicate {
  readonly type: PredicateKind.TagTitleExact;
  readonly title: string;
}

export interface TsQueryPredicate {
  readonly type: PredicateKind.TsQuery;
  readonly tsQuery: string;
}

export interface NoOpPredicate {
  readonly type: PredicateKind.NoOp;
}

export interface NotPredicate {
  readonly type: PredicateKind.Not;
  readonly predicate: Predicate;
}

export interface ResultTypePredicate {
  readonly type: PredicateKind.ResultType;
  readonly resultType: "tag" | "insight" | "highlight" | "note";
}

export type Predicate =
  | AndPredicate
  | NoOpPredicate
  | OrPredicate
  | ProjectTitleExactPredicate
  | ProjectTitleFuzzyPredicate
  | ProjectIdPredicate
  | TagIdPredicate
  | TagTitleFuzzyPredicate
  | TagTitleExactPredicate
  | TsQueryPredicate
  | NotPredicate
  | ResultTypePredicate;

export function refactorTypeTo(node: QueryNode, text: string, type: ResultTypePredicate["resultType"] | null): QueryNode {
  const children = [];

  for (const child of node.children) {
    let fieldCandidate = child;
    if (child instanceof NegateNode) {
      fieldCandidate = child.expression;
    }
    if (fieldCandidate instanceof FieldNode) {
      if (fieldCandidate.name.text(text) === "type") {
        continue;
      }
    }
    children.push(child);
  }

  if (type !== null) {
    // If the node that will follow the type:___ field starts with
    // non-whitespace, insert some whitespace to separate it from the `type:___`
    // field.
    if (getFullText(children[0], text).match(/^[^\s]/) !== null) {
      children.splice(0, 0, new InsertedToken(" "));
    }
    children.splice(0, 0, new InsertedToken(`type:${type}`));
  }

  return new QueryNode(children);
}
